// Libraries
import { Injectable } from '@angular/core';
import { Store } from '@ngrx/store';
import { Actions, Effect, ofType } from '@ngrx/effects';
import { Observable, of } from 'rxjs';
import { mergeMap, withLatestFrom, map, filter } from 'rxjs/operators';

// Models
import { Note } from '@app/core/models/note';
import { DegreePlan } from '@app/core/models/degree-plan';

// Services
import { DegreePlannerApiService } from '@app/degree-planner/services/api.service';

// State management
import {
  NoteActionTypes,
  WriteNoteRequest,
  WriteNoteResponse,
  DeleteNoteRequest,
  DeleteNoteResponse,
} from '@app/degree-planner/store/actions/note.actions';
import { getDegreePlannerState } from '@app/degree-planner/store/selectors';
import { GlobalState } from '@app/core/state';
import { DegreePlannerState } from '@app/degree-planner/store/state';

@Injectable()
export class NoteEffects {
  constructor(
    private actions$: Actions,
    private store$: Store<GlobalState>,
    private api: DegreePlannerApiService,
  ) {}

  @Effect()
  writeNote$: Observable<WriteNoteResponse> = this.actions$.pipe(
    ofType<WriteNoteRequest>(NoteActionTypes.WriteNoteRequest),

    // Get the most recent Degree Planner state object from the store. This is
    // used to decide to fire either the `updateNote` API or `createNote` API.
    withLatestFrom(this.store$.select(getDegreePlannerState)),

    // Only handle WriteNote actions when a current plan ID is set.
    filter(([_, state]) => typeof state.visibleDegreePlan !== undefined),

    // Using the action and State objects, determine whether to fire the
    // `updateNote` or `createNote` API. Both of these API calls return
    // an observable wrapper around the modified/created Note object.
    mergeMap(([action, state]) => {
      const planId = (state.visibleDegreePlan as DegreePlan).roadmapId;
      const termCode = action.payload.termCode;
      const noteText = action.payload.noteText;
      const existingNote = getExistingNote(state, termCode);
      if (existingNote !== undefined) {
        // Since the term DOES have a note, update the existing note
        const noteId = existingNote.id;
        return this.api.updateNote(planId, termCode, noteText, noteId);
      } else {
        // Since the term DOES NOT have a note, create a new note
        return this.api.createNote(planId, termCode, noteText);
      }
    }),

    // Dispatch an `WriteNoteSuccess` action so that the State
    // object can be updated with the new Note data.
    map(updatedNote => new WriteNoteResponse({ updatedNote })),
  );

  @Effect()
  deleteNote$: Observable<DeleteNoteResponse> = this.actions$.pipe(
    ofType<DeleteNoteRequest>(NoteActionTypes.DeleteNoteRequest),

    // Get the most recent Degree Planner state object.
    // This is used to lookup the Note ID.
    withLatestFrom(this.store$.select(getDegreePlannerState)),

    // Only handle DeleteNote actions when a current plan ID is set.
    filter(([_, state]) => typeof state.visibleDegreePlan !== undefined),

    // Using the action and State objects, fire the `deleteNote` API.
    mergeMap(([action, state]) => {
      const planId = (state.visibleDegreePlan as DegreePlan).roadmapId;
      const termCode = action.payload.termCode;
      const existingNote = getExistingNote(state, termCode);
      if (existingNote !== undefined) {
        // Since the term DOES have a note, delete it
        const noteId = existingNote.id;
        return this.api.deleteNote(planId, noteId).pipe(map(() => termCode));
      } else {
        // Since the term DID NOT already have a note, return an action that
        // makes no changes to the current state
        return of(termCode);
      }
    }),

    // Dispatch an `DeleteNoteSuccess` action so that the
    // State object can be updated with the note removed.
    map(termCode => new DeleteNoteResponse({ termCode })),
  );
}

const getExistingNote = (state: DegreePlannerState, termCode: string) => {
  for (const term of state.visibleTerms || []) {
    if (term.termCode === termCode) {
      return term.note;
    }
  }
  return undefined;
};
